from ..Base.SignalInterface import Signal
from scipy.fftpack import fftfreq
from scipy.fftpack import fft
from scipy.fftpack import ifft
import numpy as np

class LowPassFilter(object):

    @staticmethod
    def inplace_ideal_lowpass(signal, pos_cutoff_frquence, neg_cutoff_frequence, fs_in_fiber=None):
        '''

        :param signal: in place filter
        :return:
        '''
        samples = LowPassFilter.ideal_lowpass(signal, pos_cutoff_frquence, neg_cutoff_frequence, fs_in_fiber)

        if fs_in_fiber is None:
            # means signal is an signal object

            pol_number = signal.pol_number
        else:
            # means signal is a ndarray
            pol_number = signal.shape[0]

        for i in range(pol_number):
            signal[i, :] = samples[i, :]
        # signal.data_sample[0, :] = sample_x
        # signal.data_sample[1, :] = sample_y
        return signal

    @staticmethod
    def ideal_lowpass(signal: Signal, pos_cutoff_frequence, neg_cutoff_frequence, fs_in_fiber=None):
        '''

        :param signal: the sampled sample will be returned, original sample of signal will not be changed
        :return:
        '''
        if fs_in_fiber is None:
            fs_in_fiber = signal.fs_in_fiber
            pol_number = signal.pol_number
            sample_number_in_fiber = signal.sample_number_in_fiber
        else:
            sample_number_in_fiber = signal.shape[1]
            pol_number = signal.shape[0]

        samples = np.zeros_like(signal[:])
        for i in range(pol_number):
            samples[i, :] = signal[i, :]

        # sample_x = signal[0, :]
        # sample_y = signal[1, :]
        freq = fftfreq(sample_number_in_fiber, 1 / fs_in_fiber)

        # sample_x_fourier_transform = fft(sample_x)
        # sample_y_fourier_transform = fft(sample_y)
        fft_sample = fft(samples, axis=1)
        # 超过滤波器带宽的频率点直接设置为0
        for i in range(pol_number):
            fft_sample[i, freq > pos_cutoff_frequence] = 0
            fft_sample[i, freq < neg_cutoff_frequence] = 0
        # sample_x_fourier_transform[freq > pos_cutoff_frequence ] = 0
        # sample_x_fourier_transform[freq < neg_cutoff_frequence] = 0
        # sample_y_fourier_transform[abs(freq) > self.bw] = 0
        samples = ifft(fft_sample, axis=1)
        # sample_x = ifft(sample_x_fourier_transform)
        # sample_y = ifft(sample_y_fourier_transform)
        return samples


    @staticmethod
    def bessel_filter(signal:Signal):
        pass




import numpy as np
import numba
from scipy.fftpack import fft, ifft, fftfreq
from scipy.signal import fftconvolve
from scipy.signal import correlate
from .dsp_tools import segment_axis, decision, cal_symbols_square_qam, cal_scaling_factor_qam, get_power, \
    cal_symbols_qam
from ..Base.SignalInterface import Signal
from ..Instrument.ElectricInstrument import PulseShaping
import os


class MatchedFilter(object):

    def __init__(self, roll_off, span, sps):
        self.h = PulseShaping.rcosdesign(int(span * sps), roll_off, 1, int(sps))
        self.delay = int(span / 2 * sps)

    def match_filter(self, signal):
        x = signal.data_sample_in_fiber[0, :]
        y = signal.data_sample_in_fiber[1, :]
        x = fftconvolve(x, self.h)
        y = fftconvolve(y, self.h)

        x = np.roll(x, -self.delay)
        y = np.roll(y, -self.delay)
        x = x[:signal.sample_number_in_fiber]
        y = y[:signal.sample_number_in_fiber]
        return x, y

    def inplace_match_filter(self, signal):
        x, y = self.match_filter(signal)
        signal.data_sample_in_fiber = np.array([x, y])
        return signal

    def __call__(self, signal):
        self.inplace_match_filter(signal)

        return signal


def syncsignal(symbol_tx, sample_rx, sps):
    '''

        :param symbol_tx: 发送符号
        :param sample_rx: 接收符号，会相对于发送符号而言存在滞后
        :param sps: samples per symbol
        :return: 收端符号移位之后的结果

        # 不会改变原信号

    '''
    assert sample_rx.ndim == 1
    assert symbol_tx.ndim == 1
    assert len(sample_rx) >= len(symbol_tx)
    symbol_tx = np.atleast_2d(symbol_tx)[0, :]
    sample_rx = np.atleast_2d(sample_rx)[0, :]

    res = correlate(sample_rx[::sps], symbol_tx)

    index = np.argmax(np.abs(res))

    out = np.roll(sample_rx, sps * (-index - 1 + symbol_tx.shape[0]))
    return out


def cd_compensation(signal: Signal, spans, inplace=False):
    '''

    This function is used for chromatic dispersion compensation in frequency domain.
    The signal is Signal object, and a new sample is created from property data_sample_in_fiber

    :param signal: Signal object
    :param spans: Span object, the signal's has propagated through these spans
    :param inplace: if True, the compensated sample will replace the original sample in signal,or new ndarray will be r
    eturned

    :return: if inplace is True, the signal object will be returned; if false the ndarray will be returned
    '''
    center_wavelength = signal.center_wave_length
    freq_vector = fftfreq(signal[0, :].shape[0], 1 / signal.fs_in_fiber)
    omeg_vector = 2 * np.pi * freq_vector

    sample = np.zeros_like(signal[:])

    for i in range(signal.shape[0]):
        sample[i, :] = signal[i, :]

    if not isinstance(spans, list):
        spans = [spans]

    for span in spans:
        beta2 = -span.beta2(center_wavelength)
        dispersion = (-1j / 2) * beta2 * omeg_vector ** 2 * span.length
        for pol_index in range(sample.shape[0]):
            sample[pol_index, :] = ifft(fft(sample[pol_index, :]) * np.exp(dispersion))

    if inplace:
        signal[:] = sample
        return signal
    else:
        return sample


def super_scalar():
    '''
        implementing superscalar algorithm to recover signal's phase
    :return:
    '''
    pass


def __pll():
    '''
        using in super scalar, should not be called outside this file
    :return:
    '''
    pass


def __ml():
    '''
        using in super scalar, should not be called outside this file
    :return:
    '''
    pass


def over_lap_save(signal, h):
    '''

    :param signal: 1d-array
    :param h: filter's taps
    implement linear convolution using over-lap-save method
    :return:
    '''
    pass


def dual_frequency_lms_equalizer_block(signal, ntaps, sps, constl=None, train_symbol=None, mu=0.001, niter=3):
    pass


@numba.jit(cache=True)
def dual_pol_time_domain_lms_equalizer_pll(signal, ntaps, sps, constl=None, train_symbol=None, mu=0.001,
                                           niter=3, g=0.01, symbol_length=65536):
    weight = __dual_pol_init_lms_weight(ntaps)  # xx xy yx yy
    data_sample_in_fiber = signal[:]
    samplex = segment_axis(data_sample_in_fiber[0, :], ntaps, ntaps - sps)
    sampley = segment_axis(data_sample_in_fiber[1, :], ntaps, ntaps - sps)
    if train_symbol is not None:
        xsymbol = train_symbol[0, ntaps // 2 // sps:]
        ysymbol = train_symbol[1, ntaps // 2 // sps:]
        # xsymbol = np.roll(xsymbol,)

    xsymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    ysymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    errorsx = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)
    errorsy = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)
    pll_phase = np.zeros((2, samplex.shape[0]), dtype=np.complex128)
    pll_error = np.zeros((2, samplex.shape[0]), dtype=np.complex128)

    hxx = np.zeros((1, samplex.shape[0]),dtype = np.float32)
    hxy = np.zeros((1, samplex.shape[0]),dtype = np.float32)
    hyx = np.zeros((1, samplex.shape[0]),dtype = np.float32)
    hyy = np.zeros((1, samplex.shape[0]),dtype = np.float32)

    for j in range(niter):
        for i in range(samplex.shape[0]):
            xout = np.sum(samplex[i, ::-1] * weight[0][0]) + np.sum(sampley[i, ::-1] * weight[1][0])
            yout = np.sum(samplex[i, ::-1] * weight[2][0]) + np.sum(sampley[i, ::-1] * weight[3][0])
            xout_cpr = xout * np.exp(-1j * pll_phase[0, i])
            yout_cpr = yout * np.exp(-1j * pll_phase[1, i])

            pll_error[0, i] = (xout * np.conj(xout_cpr)).imag
            pll_error[1, i] = (yout * np.conj(yout_cpr)).imag

            pll_phase[0, i] = g * pll_error[0, i] + pll_phase[0, i]

            pll_phase[1, i] = g * pll_error[1, i] + pll_phase[1, i]
            #  is not None and j == 0:
            if train_symbol is not None:
                errorx = __calc_error_train(xout_cpr, symbol=xsymbol[i])
                errory = __calc_error_train(yout_cpr, symbol=ysymbol[i])
            else:
                assert constl is not None
                # assert constl.ndim ==1

                if constl.ndim != 2:
                    raise Exception("constl must be 2d")

                errorx = __calc_error_dd(xout_cpr, constl)
                errory = __calc_error_dd(yout_cpr, constl)

            weight[0][0] = weight[0][0] + mu * errorx * np.conj(samplex[i, ::-1])
            weight[1][0] = weight[1][0] + mu * errorx * np.conj(sampley[i, ::-1])

            weight[2][0] = weight[2][0] + mu * errory * np.conj(samplex[i, ::-1])
            weight[3][0] = weight[3][0] + mu * errory * np.conj(sampley[i, ::-1])

            if j == niter - 1:
                xsymbols[0, i] = xout_cpr
                ysymbols[0, i] = yout_cpr
                hxx[0,i] = weight[0][0,ntaps//2+1].real
                hxy[0,i] = weight[1][0,ntaps//2+1].real

                hyx[0,i] = weight[2][0,ntaps//2+1].real

                hyy[0, i] = weight[3][0, ntaps//2+1].real

            errorsx[0, i] = np.abs(errorx)
            errorsy[0, i] = np.abs(errory)

    return xsymbols[0, :], ysymbols[0, :], weight, errorsx, errorsy,(hxx,hxy,hyx,hyy)


@numba.jit(cache=True)
def dual_pol_time_domain_lms_equalizer(signal, ntaps, sps, constl=None, train_symbol=None, mu=0.001,
                                       niter=3, symbol_length=65536):
    '''

    :param signal:         signal to be equalized, a numpy array, 2d-array,each row represent a polarizaiton
    :param ntaps:          the tap of the filter
    :param sps:            sample per symbol
    :param constl:         the constellation, used in dd
    :param train_symbol:   symbol to train
    :param mu:             the learning rate, default to 0.001
    :param niter:          the number of equlizeing operation
    :return: equalized symbols, errors
    '''
    data_sample_in_fiber = signal[:]
    samplex = segment_axis(data_sample_in_fiber[0, :], ntaps, ntaps - sps)
    sampley = segment_axis(data_sample_in_fiber[1, :], ntaps, ntaps - sps)

    if train_symbol is not None:
        xsymbol = train_symbol[0, ntaps // 2 // sps:]
        ysymbol = train_symbol[1, ntaps // 2 // sps:]
        # xsymbol = np.roll(xsymbol,)

    weight = __dual_pol_init_lms_weight(ntaps)  # xx xy yx yy
    xsymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    ysymbols = np.zeros((1, symbol_length), dtype=np.complex128)
    errorsx = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)
    errorsy = np.zeros((1, niter * samplex.shape[0]), dtype=np.float64)

    hxx = np.zeros((1,samplex.shape[0]))
    hxy = np.zeros((1, samplex.shape[0]))
    hyx = np.zeros((1, samplex.shape[0]))
    hyy = np.zeros((1, samplex.shape[0]))

    for j in range(niter):
        for i in range(samplex.shape[0]):
            xout = np.sum(samplex[i, ::-1] * weight[0][0]) + np.sum(sampley[i, ::-1] * weight[1][0])
            yout = np.sum(samplex[i, ::-1] * weight[2][0]) + np.sum(sampley[i, ::-1] * weight[3][0])

            if train_symbol is not None and j == 0:
                errorx = __calc_error_train(xout, symbol=xsymbol[i])
                errory = __calc_error_train(yout, symbol=ysymbol[i])
            else:
                assert constl is not None
                # assert constl.ndim ==1

                if constl.ndim != 2:
                    raise Exception("constl must be 2d")

                errorx = __calc_error_dd(xout, constl)
                errory = __calc_error_dd(yout, constl)

            weight[0][0] = weight[0][0] + mu * errorx * np.conj(samplex[i, ::-1])
            weight[1][0] = weight[1][0] + mu * errorx * np.conj(sampley[i, ::-1])

            weight[2][0] = weight[2][0] + mu * errory * np.conj(samplex[i, ::-1])
            weight[3][0] = weight[3][0] + mu * errory * np.conj(sampley[i, ::-1])

            if j == niter - 1:
                xsymbols[0, i] = xout
                ysymbols[0, i] = yout
            errorsx[0, i] = np.abs(errorx)
            errorsy[0, i] = np.abs(errory)

    return xsymbols, ysymbols, weight, errorsx, errorsy


@numba.jit(cache=True)
def __calc_error_dd(xout, constl):
    '''
        only use in dual_pol_time_domain_lms_equalizer, should not be used outside this file
        xout: a symbol to be decision,one element
        constl: must be 2d array
    '''
    symbol = decision(xout, constl)
    return symbol - xout


@numba.jit(cache=True)
def __calc_error_train(xout, symbol):
    '''
        only use in dual_pol_time_domain_lms_equalizer, should not be used outside this file
        xout: symbol in ,one element
        symbol: True symbol,one element
    '''
    error = symbol - xout
    return error


@numba.jit(cache=True)
def __dual_pol_init_lms_weight(ntaps):
    '''
     only use in dual_pol_time_domain_lms_equalizer, should not be used outside this file
     implement filter taps initialization

    :param ntaps:
    :return:
    '''
    # hxx = np.zeros((1, ntaps), dtype=np.complex128)
    #
    # hxy = np.zeros((1, ntaps), dtype=np.complex128)
    # hyx = np.zeros((1, ntaps), dtype=np.complex128)
    # hyy = np.zeros((1, ntaps), dtype=np.complex128)
    #
    # hxx[0, ntaps // 2] = 1
    # hyy[0, ntaps // 2] = 1

    hxx = np.zeros((1, ntaps), dtype=np.complex128)

    hxy = np.zeros((1, ntaps), dtype=np.complex128)
    hyx = np.zeros((1, ntaps), dtype=np.complex128)
    hyy = np.zeros((1, ntaps), dtype=np.complex128)

    hyx[0, ntaps // 2] = 1
    hxy[0, ntaps // 2] = 1
    return (hxx, hxy, hyx, hyy)


def demap_to_msg(rx_symbols, order, do_normal=True):
    '''

    :param rx_symbols:1d array or 2d array if 2d the shape[0] must be 1
    :param do_normal: if True the rx_symbols will be normalized to 1
    :return: 2d array msg, the shape[0] is 1
    '''
    from scipy.io import loadmat

    base_path = os.path.abspath(__file__)
    base_path = os.path.dirname(os.path.dirname(base_path))

    if order == 8:
        raise NotImplementedError('Not implemented yet')

    constl = cal_symbols_qam(order) / np.sqrt(cal_scaling_factor_qam(order))
    rx_symbols = np.atleast_2d(rx_symbols)
    assert rx_symbols.shape[0] == 1
    rx_symbols = rx_symbols[0]

    qam_data = loadmat(f'{base_path}/qamdata/{order}qam.mat')['x'][0]
    msg = np.zeros((1, rx_symbols.shape[0]))
    if do_normal:
        rx_symbols = rx_symbols / np.sqrt(np.mean(rx_symbols.imag ** 2 + rx_symbols.real ** 2))

    for index, symbol in enumerate(rx_symbols):
        decision_symbol = decision(symbol, constl=np.atleast_2d(constl))

        choose = np.abs(decision_symbol - qam_data) < np.spacing(1)
        #         print(choose)
        #         print(np.nonzero(choose)[0])
        msg[0, index] = (np.nonzero(choose)[0])
    #         print(msg[0,index])
    #         break

    return msg.astype(np.int)


def main():
    import matplotlib.pyplot as plt
    from scipy.io import loadmat

    cc = loadmat('test2.mat')
    tx = cc['txsymbols']
    sam = cc['rxSignalIn']

    import time

    now = time.time()

    constl = cal_symbols_qam(16)
    constl = constl / np.sqrt(cal_scaling_factor_qam(16))
    constl = np.atleast_2d(constl)
    xsym, ysym, wer, errorsx, errorsy = dual_pol_time_domain_lms_equalizer(sam[:], 13, 2, constl=constl,
                                                                           train_symbol=tx,
                                                                           mu=0.001, niter=3)
    plt.plot(errorsx)
    print(time.time() - now)


if __name__ == '__main__':
    main()



from .ElectricInstrument import AWG
from .ElectricInstrument import ADC
from .ElectricInstrument import PulseShaping
from .ElectricInstrument import DAC
from .OpticalInstrument import Edfa
from .OpticalInstrument import Multiplex
from .OpticalInstrument import Demultiplex,WSS



from numpy.fft import fftfreq
from scipy.fftpack import fft, ifft

try:
    import cupy as cp
    from cupy.fft import fft as cfft
    from cupy.fft import ifft as icfft
except Exception as e:
    print('cupy can not be used')

from ..Base import SignalInterface
import numpy as np
from scipy.constants import c
import math
# import progressbar


# import arrayfire as af


class Awgn(object):

    def __init__(self, signal_power, request_snr, sps):
        self.signal_power = signal_power
        self.request_snr = request_snr
        self.sps = sps

    def __call__(self, shape):
        noise_power = self.signal_power / self.request_snr_lin

        noise_power = noise_power * self.sps
        noise_x = np.sqrt(noise_power / 2 / 2) * (
                np.random.randn(shape) + 1j * np.random.randn(shape))

        noise_y = np.sqrt(noise_power / 2 / 2) * (
                np.random.randn(shape) + 1j * np.random.randn(shape))

        return noise_x, noise_y

    @property
    def request_snr_lin(self):
        return 10 ** (self.request_snr / 10)


class LinearFiber(object):
    '''
        property:
            self.alpha  [db/km]
            self.D [ps/nm/km]
            self.length [km]
            self.wave_length:[nm]
            self.beta2: caculate beta2 from D,s^2/km
            self.slope: derivative of self.D ps/nm^2/km
            self.beta3_reference: s^3/km
        method:
            __call__: the signal will
    '''

    def __init__(self, alpha, D, length, slope=0, reference_wave_length=1550):
        self.alpha = alpha
        self.D = D
        self.length = length
        self.reference_wave_length = reference_wave_length
        self.slope = slope

    @property
    def beta3_reference(self):
        res = (self.reference_wave_length * 1e-12 / 2 / np.pi / c / 1e-3) ** 2 * (
                2 * self.reference_wave_length * 1e-12 * self.D + (
                self.reference_wave_length * 1e-12) ** 2 * self.slope * 1e12)

        return res

    @property
    def alpha_lin(self):
        # [1/km]
        return 0.1 * self.alpha * np.log(10)

    def leff(self, length):
        '''

        :param length: the length of a fiber [km]
        :return: the effective length [km]
        '''
        effective_length = 1 - np.exp(-self.alpha_lin * length)
        effective_length = effective_length / self.alpha_lin
        return effective_length

    @property
    def beta2_reference(self):
        return -self.D * (self.reference_wave_length * 1e-12) ** 2 / 2 / np.pi / c / 1e-3

    def beta2(self, wave_length):
        '''

        :param wave_length: [m]
        :return: beta2 at wave_length [s^2/km]
        '''
        dw = 2 * np.pi * c * (1 / wave_length - 1 / (self.reference_wave_length * 1e-9))
        return self.beta2_reference + self.beta3_reference * dw

    def _prop(self, signal: SignalInterface.Signal):
        '''

        :param signal: signal object to propagation across this fiber
        :return: ndarray
        '''
        center_lambda = signal.center_wave_length

        after_prop = np.zeros_like(signal[:])
        for pol in range(0, signal.pol_number):
            sample = signal[pol, :]
            sample_fft = fft(sample)
            freq = fftfreq(signal.sample_number_in_fiber, 1 / signal.fs_in_fiber)
            omeg = 2 * np.pi * freq

            after_prop[pol, :] = sample_fft * np.exp(-self.alpha_lin * self.length / 2)
            after_prop[pol, :] = ifft(after_prop[pol, :])

            disp = np.exp(-1j / 2 * self.beta2(center_lambda) * omeg ** 2 * self.length)
            after_prop[pol, :] = ifft(fft(after_prop[pol, :]) * disp)

        return np.atleast_2d(after_prop)

    def inplace_prop(self, signal: SignalInterface.Signal):
        after_prop = self._prop(signal)

        for i in range(signal.pol_number):
            signal[i, :] = after_prop[i, :]
        return signal

    def __call__(self, signal):
        self.inplace_prop(signal)
        return signal

    def __str__(self):
        string = f"alpha is {self.alpha} [db/km]\n" \
            f"beta2 is {self.beta2_reference} [s^2/km]\n" \
            f"beta3 is {self.beta3_reference} []\n" \
            f"D is {self.D} ps/nm/km\n" \
            f"length is {self.length} km\n" \
            f"reference wave length is {self.reference_wave_length * 1e9} [nm]"

        return string

    def __repr__(self):
        return self.__str__()


class NonlinearFiber(LinearFiber):

    def __init__(self, alpha, D, length, gamma, slope=0, step_length=5 / 1000, reference_wave_length=1550):
        super().__init__(alpha, D, length, slope=slope, reference_wave_length=reference_wave_length)
        self.gamma = gamma
        self.step_length = step_length

    @property
    def step_length_eff(self):
        return (1 - np.exp(-self.alpha_lin * self.step_length)) / self.alpha_lin

    def cupy_prop(self, signal):
#         print("cupy is used")

        if signal.pol_number != 2:
            raise Exception("only dp signal supported at this time")
        step_number = self.length / self.step_length
        step_number = int(np.floor(step_number))
        temp = np.zeros_like(signal[:])
        freq = fftfreq(signal.sample_number_in_fiber, 1 / signal.fs_in_fiber)
        freq_gpu = cp.asarray(freq)
        omeg = 2 * np.pi * freq_gpu
        D = -1j / 2 * self.beta2(signal.center_wave_length) * omeg ** 2
        N = 8 / 9 * 1j * self.gamma
        atten = -self.alpha_lin / 2

        time_x = cp.asarray(signal[0, :])
        time_y = cp.asarray(signal[1, :])

#         print('*' * 20 + "begin simulation" + '*' * 20)
#         bar = progressbar.ProgressBar(max_value=step_number)
        for i in range(step_number):
#             bar.update(i + 1)
            fftx = cfft(time_x)
            ffty = cfft(time_y)

            time_x, time_y = self.linear_prop_cupy(D, fftx, ffty, self.step_length / 2)
            time_x, time_y = self.nonlinear_prop_cupy(N, time_x, time_y)
            time_x = time_x * math.exp(atten * self.step_length)
            time_y = time_y * math.exp(atten * self.step_length)

            fftx = cfft(time_x)
            ffty = cfft(time_y)

            time_x, time_y = self.linear_prop_cupy(D, fftx, ffty, self.step_length / 2)

#         bar.finish()

#         print('*' * 20 + "end simulation" + '*' * 20)

        last_step = self.length - self.step_length * step_number
        last_step_eff = (1 - np.exp(-self.alpha_lin * last_step)) / self.alpha_lin
        if last_step == 0:
            time_x = cp.asnumpy(time_x)
            time_y = cp.asnumpy(time_y)
            temp[0, :] = time_x
            temp[1, :] = time_y
            return temp

        fftx = cfft(time_x)
        ffty = cfft(time_y)

        time_x, time_y = self.linear_prop_cupy(D, fftx, ffty, last_step / 2)
        time_x, time_y = self.nonlinear_prop_cupy(N, time_x, time_y, last_step_eff)
        time_x = time_x * math.exp(atten * last_step)
        time_y = time_y * math.exp(atten * last_step)

        fftx = cfft(time_x)
        ffty = cfft(time_y)
        time_x, time_y = self.linear_prop_cupy(D, fftx, ffty, last_step / 2)

        temp[0, :] = cp.asnumpy(time_x)
        temp[1, :] = cp.asnumpy(time_y)
        return temp

    def nonlinear_prop_cupy(self, N, time_x, time_y, step_length=None):

        if step_length is None:
            time_x = time_x * cp.exp(
                N * self.step_length_eff * (cp.abs(time_x) ** 2 + cp.abs(
                    time_y) ** 2))
            time_y = time_y * cp.exp(
                N * self.step_length_eff * (cp.abs(time_x) ** 2 + cp.abs(time_y) ** 2))
        else:
            time_x = time_x * cp.exp(
                N * step_length * (cp.abs(time_x) ** 2 + cp.abs(
                    time_y) ** 2))
            time_y = time_y * cp.exp(
                N * step_length * (cp.abs(time_x) ** 2 + cp.abs(time_y) ** 2))

        return time_x, time_y

    def linear_prop_cupy(self, D, fftx, ffty, length):
        time_x = icfft(fftx * cp.exp(D * length))
        time_y = icfft(ffty * cp.exp(D * length))

        return time_x, time_y

    def inplace_prop(self, signal):
        after_prop = self.cupy_prop(signal)
        signal.data_sample_in_fiber = after_prop
        return signal

    def __call__(self, signal):
        self.inplace_prop(signal)
        return signal

    def __str__(self):
        string = super(NonlinearFiber, self).__str__()
        string = f"{string}" \
            f"the step length is {self.step_length} [km]\n"
        return string

    def __repr__(self):
        return self.__str__()


class NonlinearFiberWithPMD(LinearFiber):

    def __init__(self, alpha, D, length, gamma, reference_length=1550, step_length=5 / 1000, is_dgd=True, **kwargs):
        '''

        :param alpha: [db/km]
        :param D: [ps/nm/km]
        :param length: [km]
        :param gamma: []
        :param reference_length:
        :param is_vector:
        :param kwargs:
                if pmd, dgd_coef should be specified,ps/sqrt(km)
                step_length [km]  the step length of split step fourier method
        '''
        super(NonlinearFiberWithPMD, self).__init__(alpha, D, length, reference_length)
        self.gamma = gamma
        if is_dgd:
            try:
                self.dgd_coef = kwargs['dgd_coef']
            except KeyError as e:
                self.dgd_coef = 0.1
        else:
            self.dgd_coef = 0

        self.step_length = step_length
        self.nplates = self.length / self.step_length
        self.nplates = int(np.floor(self.nplates))

        if self.nplates * self.step_length < self.length:
            self.last_step = self.length - self.nplates * self.step_length
        else:
            self.last_step = 0

        self.dgd = self.dgd_coef * np.sqrt(self.step_length * self.nplates) * 1e-12  # s

        if self.last_step != 0:
            self.last_dgd = self.dgd_coef * np.sqrt(self.last_step)
        else:
            self.last_dgd = 0

        self.dgdrms = np.sqrt((3 * np.pi) / 8) * self.dgd / np.sqrt(self.nplates)

        if self.last_dgd != 0:
            self.dgdrms_last = np.sqrt((3 * np.pi) / 8) * self.last_dgd / np.sqrt(1)
        else:
            self.dgdrms_last = 0

    @property
    def betat(self):
        return 1

    def vector_prop(self, signal: SignalInterface.WdmSignal):
        print("The Mankov equation of  will be solved using split step fourier method")
        sample_x = signal[0, :]
        sample_y = signal[1, :]

        sample_x_gpu = cp.asarray(sample_x)
        sample_y_gpu = cp.asarray(sample_y)

        wave_length = signal.center_wave_length

        beta2 = self.beta2(wave_length=wave_length)
        freq = fftfreq(signal.sample_number_in_fiber, 1 / signal.fs_in_fiber)
        omeg = 2 * np.pi * freq
        betat = -0.5 * omeg ** 2 * beta2
        db1 = self.dgdrms * omeg
        db1_last = self.dgdrms_last * omeg

        sample_x_gpu, sample_y_gpu = self.__vector_prop(sample_x_gpu, sample_y_gpu, betat, self.gamma, db1, db1_last)
        sample_x_cpu = cp.asnumpy(sample_x_gpu)
        sample_y_cpu = cp.asnumpy(sample_y_gpu)
        return sample_x_cpu, sample_y_cpu

    def __call__(self, signal):

        signal[0, :], signal[1, :] = self.vector_prop(signal)
        return signal

    def __vector_prop(self, ux, uy, betat, gamma, db1, db1_last):
        betat = cp.asarray(betat)
        db1 = cp.asarray(db1)
        db1_last = cp.asarray(db1_last)

        gamma = gamma * 8 / 9
        halfalpha = 0.5 * self.alpha_lin
        # bar = progressbar.ProgressBar(max_value=self.nplates)
        for i in range(self.nplates):
            ux, uy = self.__vector_nonlinear_step(ux, uy, gamma, self.step_length)
            ux, uy = self.__vector_linear_step(ux, uy, betat, db1, self.step_length)
            ux = ux * math.exp(-halfalpha * self.step_length)
            uy = uy * math.exp(-halfalpha * self.step_length)
            # bar.update(i + 1)

        if self.last_step:
            print(f"the last step {self.last_step}m is going to be propagated")
            ux, uy = self.__vector_nonlinear_step(ux, uy, gamma, self.last_step)
            ux, uy = self.__vector_linear_step(ux, uy, betat, db1_last, self.last_step)
            ux = ux * math.exp(-halfalpha * self.last_step)
            uy = uy * math.exp(-halfalpha * self.last_step)

        return ux, uy

    def __vector_linear_step(self, ux, uy, betat, db1, step_length):
        ux = cfft(ux)
        uy = cfft(uy)
        phi = np.random.rand(1, 1) * np.pi - 0.5 * np.pi
        kapa = 0.5 * np.arcsin(2 * np.random.rand(1, 1) - 1)

        rotation_matrix = np.array([[np.cos(phi) * np.cos(kapa) + 1j * np.sin(phi) * np.sin(kapa),
                                     np.sin(phi) * np.cos(kapa) - 1j * np.cos(phi) * np.sin(kapa)],
                                    [-np.sin(phi) * np.cos(kapa) - 1j * np.cos(phi) * np.sin(kapa),
                                     np.cos(phi) * np.cos(kapa) - 1j * np.sin(phi) * np.sin(kapa)]])

        rotation_matrix_conj = rotation_matrix.T.conj()
        # rotate to principle state to of the fiber
        uux = rotation_matrix[0, 0] * ux + rotation_matrix[0, 1] * uy
        uuy = rotation_matrix[1, 0] * ux + rotation_matrix[1, 1] * uy

        # betat = -1/2*beta2*omeg**2
        disp = betat * step_length
        #
        linear_operator_xpol = (0.5 * db1 + disp) * 1j
        linear_operator_ypol = (disp - 0.5 * db1) * 1j
        # linear_operator_xpol = disp * 1j
        # linear_operator_ypol = disp * 1j

        uux = cp.exp(linear_operator_xpol) * uux
        uuy = cp.exp(linear_operator_ypol) * uuy

        ux = rotation_matrix_conj[0, 0] * uux + rotation_matrix_conj[0, 1] * uuy
        uy = rotation_matrix_conj[1, 0] * uux + rotation_matrix_conj[1, 1] * uuy

        ux = cp.ifft(ux)
        uy = cp.ifft(uy)

        return ux, uy

    def __vector_nonlinear_step(self, ux, uy, gamma_mankorv, step_length):
        '''

        :param ux: samples of x-pol
        :param uy: samples of y-pol
        :param gamma_mankorv: 8/9*gamma，represent average over 那个啥球上
        :return: samples after nonlinear propagation
        '''
        leff = self.leff(step_length)
        power = cp.abs(ux) ** 2 + cp.abs(uy) ** 2
        gamma_mankorv = gamma_mankorv * leff
        ux = ux * cp.exp(1j * gamma_mankorv * power)
        uy = uy * cp.exp(1j * gamma_mankorv * power)
        return ux, uy

    def __str__(self):
        string = super().__str__()
        string = f'{string}\n' \
            f'step length for split fourier method is {self.step_length}'

        return string

    def __repr__(self):
        return self.__str__()


if __name__ == "__main__":
    pass




from setuptools import setup, find_packages

setup(
    name="ocspy",
    version="0.1.3",
    keywords=("pip", "optical communication",
              "coherent optical", "ocspy", "ocspy"),
    description="simulate optical communication",
    long_description="simulate optical communication",
    license="MIT Licence",
    packages=find_packages(),

    url="",
    author="nigulasikaochuan",
    author_email="nigulasikaochuan@sjtu.edu.cn",


    include_package_data=True,
    platforms="any",
    install_requires=[]
)
